@@ -8,10 +8,10 @@ ROCM_PATH=${ROCM_PATH:="/opt/rocm"}
 
 # Default values
 build_address_sanitizer=false
-build_allreduce_only=false
-build_bfd=true
-build_freorg_bkwdcomp=true
+build_bfd=false
+build_freorg_bkwdcomp=false
 build_local_gpu_only=false
+build_amdgpu_targets=""
 build_package=false
 build_release=true
 build_static=false
@@ -22,7 +22,8 @@ collective_trace=true
 enable_ninja=""
 install_dependencies=false
 install_library=false
-num_parallel_jobs=16
+msccl_kernel_enabled=true
+num_parallel_jobs=$(nproc)
 npkit_enabled=false
 run_tests=false
 run_tests_all=false
@@ -36,16 +37,17 @@ function display_help()
     echo "RCCL build & installation helper script"
     echo " Options:"
     echo "       --address-sanitizer     Build with address sanitizer enabled"
-    echo "       --build_allreduce_only  Build only AllReduce + sum + float kernel"
     echo "    -d|--dependencies          Install RCCL depdencencies"
     echo "       --debug                 Build debug library"
-    echo "       --disable_backtrace     Build without custom backtrace support"
+    echo "       --enable_backtrace      Build with custom backtrace support"
     echo "       --disable-colltrace     Build without collective trace"
+    echo "       --disable-msccl-kernel  Build without MSCCL kernels"
     echo "    -f|--fast                  Quick-build RCCL (local gpu arch only, no backtrace, and collective trace support)"
     echo "    -h|--help                  Prints this help message"
     echo "    -i|--install               Install RCCL library (see --prefix argument below)"
     echo "    -j|--jobs                  Specify how many parallel compilation jobs to run ($num_parallel_jobs by default)"
     echo "    -l|--local_gpu_only        Only compile for local GPU architecture"
+    echo "       --amdgpu_targets        Only compile for specified GPU architecture(s). For multiple targets, seperate by ';' (builds for all supported GPU architectures by default)"
     echo "       --no_clean              Don't delete files if they already exist"
     echo "       --npkit-enable          Compile with npkit enabled"
     echo "    -p|--package_build         Build RCCL package"
@@ -66,7 +68,7 @@ function display_help()
 # check if we have a modern version of getopt that can handle whitespace and long parameters
 getopt -T
 if [[ $? -eq 4 ]]; then
-    GETOPT_PARSE=$(getopt --name "${0}" --options dfhij:lprt --longoptions address-sanitizer,build_allreduce_only,dependencies,debug,disable_backtrace,disable-colltrace,fast,help,install,jobs:,local_gpu_only,no_clean,npkit-enable,package_build,prefix:,rm-legacy-include-dir,run_tests_all,run_tests_quick,tests_build,time-trace,verbose -- "$@")
+    GETOPT_PARSE=$(getopt --name "${0}" --options dfhij:lprt --longoptions address-sanitizer,dependencies,debug,enable_backtrace,disable-colltrace,disable-msccl-kernel,fast,help,install,jobs:,local_gpu_only,amdgpu_targets:,no_clean,npkit-enable,package_build,prefix:,rm-legacy-include-dir,run_tests_all,run_tests_quick,static,tests_build,time-trace,verbose -- "$@")
 else
     echo "Need a new version of getopt"
     exit 1
@@ -81,28 +83,29 @@ eval set -- "${GETOPT_PARSE}"
 
 while true; do
     case "${1}" in
-         --address-sanitizer)        build_address_sanitizer=true;                                       shift ;;
-         --build_allreduce_only)     build_allreduce_only=true;                                          shift ;;
-    -d | --dependencies)             install_dependencies=true;                                          shift ;;
-         --debug)                    build_release=false;                                                shift ;;
-         --disable_backtrace)        build_bfd=false;                                                    shift ;;
-         --disable-colltrace)        collective_trace=false;                                             shift ;;
-    -f | --fast)                     build_bfd=false; build_local_gpu_only=true; collective_trace=false; shift ;;
-    -h | --help)                     display_help;                                                       exit 0 ;;
-    -i | --install)                  install_library=true;                                               shift ;;
-    -j | --jobs)                     num_parallel_jobs=${2};                                             shift 2 ;;
-    -l | --local_gpu_only)           build_local_gpu_only=true;                                          shift ;;
-         --no_clean)                 clean_build=false;                                                  shift ;;
-         --npkit-enable)             npkit_enabled=true;                                                 shift ;;
-    -p | --package_build)            build_package=true;                                                 shift ;;
-         --prefix)                   install_prefix=${2};                                                shift 2 ;;
-         --rm-legacy-include-dir)    build_freorg_bkwdcomp=false;                                        shift ;;
-    -r | --run_tests_quick)          run_tests=true;                                                     shift ;;
-         --run_tests_all)            run_tests=true; run_tests_all=true;                                 shift ;;
-         --static)                   build_static=true;                                                  shift ;;
-    -t | --tests_build)              build_tests=true;                                                   shift ;;
-         --time-trace)               time_trace=true;                                                    shift ;;
-         --verbose)                  build_verbose=1;                                                    shift ;;
+         --address-sanitizer)        build_address_sanitizer=true;                                                                     shift ;;
+    -d | --dependencies)             install_dependencies=true;                                                                        shift ;;
+         --debug)                    build_release=false;                                                                              shift ;;
+         --enable_backtrace)         build_bfd=true;                                                                                   shift ;;
+         --disable-colltrace)        collective_trace=false;                                                                           shift ;;
+         --disable-msccl-kernel)     msccl_kernel_enabled=false;                                                                       shift ;;
+    -f | --fast)                     build_local_gpu_only=true; collective_trace=false; msccl_kernel_enabled=false;                    shift ;;
+    -h | --help)                     display_help;                                                                                     exit 0 ;;
+    -i | --install)                  install_library=true;                                                                             shift ;;
+    -j | --jobs)                     num_parallel_jobs=${2};                                                                           shift 2 ;;
+    -l | --local_gpu_only)           build_local_gpu_only=true;                                                                        shift ;;
+         --amdgpu_targets)           build_amdgpu_targets=${2};                                                                        shift 2 ;;
+         --no_clean)                 clean_build=false;                                                                                shift ;;
+         --npkit-enable)             npkit_enabled=true;                                                                               shift ;;
+    -p | --package_build)            build_package=true;                                                                               shift ;;
+         --prefix)                   install_prefix=${2};                                                                              shift 2 ;;
+         --rm-legacy-include-dir)    build_freorg_bkwdcomp=false;                                                                      shift ;;
+    -r | --run_tests_quick)          run_tests=true;                                                                                   shift ;;
+         --run_tests_all)            run_tests=true; run_tests_all=true;                                                               shift ;;
+         --static)                   build_static=true;                                                                                shift ;;
+    -t | --tests_build)              build_tests=true;                                                                                 shift ;;
+         --time-trace)               time_trace=true;                                                                                  shift ;;
+         --verbose)                  build_verbose=1;                                                                                  shift ;;
     --) shift ; break ;;
     *)  echo "Unexpected command line parameter received; aborting";
         exit 1
@@ -176,14 +179,9 @@ if [[ "${build_address_sanitizer}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DBUILD_ADDRESS_SANITIZER=ON"
 fi
 
-# AllReduce only
-if [[ "${build_allreduce_only}" == true ]]; then
-    cmake_common_options="${cmake_common_options} -DBUILD_ALLREDUCE_ONLY=ON"
-fi
-
 # Backtrace support
-if [[ "${build_bfd}" == false ]]; then
-    cmake_common_options="${cmake_common_options} -DBUILD_BFD=OFF"
+if [[ "${build_bfd}" == true ]]; then
+    cmake_common_options="${cmake_common_options} -DBUILD_BFD=ON"
 fi
 
 # Backward compatibility wrappers
@@ -198,6 +196,11 @@ if [[ "$build_local_gpu_only" == true ]]; then
     cmake_common_options="${cmake_common_options} -DBUILD_LOCAL_GPU_TARGET_ONLY=ON"
 fi
 
+# Build for specified GPU target(s) only
+if [[ ! -z "$build_amdgpu_targets" ]]; then
+    cmake_common_options="${cmake_common_options} -DAMDGPU_TARGETS=${build_amdgpu_targets}"
+fi
+
 # shared vs static
 if [[ "${build_static}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DBUILD_SHARED_LIBS=OFF"
@@ -208,6 +211,10 @@ if [[ "${collective_trace}" == false ]]; then
     cmake_common_options="${cmake_common_options} -DCOLLTRACE=OFF"
 fi
 
+if [[ "${msccl_kernel_enabled}" == false ]]; then
+    cmake_common_options="${cmake_common_options} -DENABLE_MSCCL_KERNEL=OFF"
+fi
+
 # Install dependencies
 if ($install_dependencies); then
     cmake_common_options="${cmake_common_options} -DINSTALL_DEPENDENCIES=ON"
@@ -273,6 +280,8 @@ if ($npkit_enabled); then
     -DENABLE_NPKIT_EVENT_PRIM_LL128_DATA_PROCESS_EXIT \
     -DENABLE_NPKIT_EVENT_NET_SEND_ENTRY \
     -DENABLE_NPKIT_EVENT_NET_SEND_EXIT \
+    -DENABLE_NPKIT_EVENT_NET_TEST_ENTRY \
+    -DENABLE_NPKIT_EVENT_NET_TEST_EXIT \
     -DENABLE_NPKIT_EVENT_NET_RECV_ENTRY \
     -DENABLE_NPKIT_EVENT_NET_RECV_EXIT \
     -DENABLE_NPKIT_EVENT_ALL_REDUCE_RING_SEND_ENTRY \
@@ -309,6 +318,20 @@ if ($npkit_enabled); then
     -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_RECV_COPY_SEND_EXIT \
     -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_DIRECT_RECV_ENTRY \
     -DENABLE_NPKIT_EVENT_ALL_GATHER_RING_DIRECT_RECV_EXIT \
+    -DENABLE_NPKIT_EVENT_MSCCL_GENERIC_OP_ENTRY \
+    -DENABLE_NPKIT_EVENT_MSCCL_GENERIC_OP_EXIT \
+    -DENABLE_NPKIT_EVENT_MSCCL_REDUCE_ENTRY \
+    -DENABLE_NPKIT_EVENT_MSCCL_REDUCE_EXIT \
+    -DENABLE_NPKIT_EVENT_MSCCL_SEND_ENTRY \
+    -DENABLE_NPKIT_EVENT_MSCCL_SEND_EXIT \
+    -DENABLE_NPKIT_EVENT_MSCCL_RECV_ENTRY \
+    -DENABLE_NPKIT_EVENT_MSCCL_RECV_EXIT \
+    -DENABLE_NPKIT_EVENT_MSCCL_RUN_ENTRY \
+    -DENABLE_NPKIT_EVENT_MSCCL_RUN_EXIT \
+    -DENABLE_NPKIT_EVENT_MSCCL_RECV_REDUCE_COPY_ENTRY \
+    -DENABLE_NPKIT_EVENT_MSCCL_RECV_REDUCE_COPY_EXIT \
+    -DENABLE_NPKIT_EVENT_MSCCL_INIT_ENTRY \
+    -DENABLE_NPKIT_EVENT_MSCCL_INIT_EXIT \
     -DENABLE_NPKIT_PRIM_COLLECT_DATA_PROCESS_TIME"
 fi
 
@@ -322,9 +345,9 @@ else
 fi
 
 if ($build_tests) || (($run_tests) && [[ ! -f ./test/rccl-UnitTests ]]); then
-    CXX=$ROCM_BIN_PATH/hipcc $cmake_executable $cmake_common_options -DBUILD_TESTS=ON -DNPKIT_FLAGS="${npkit_options}" -DCMAKE_INSTALL_PREFIX=$ROCM_PATH -DROCM_PATH=$ROCM_PATH $enable_ninja ../../.
+    CXX=$ROCM_BIN_PATH/hipcc $cmake_executable $cmake_common_options -DBUILD_TESTS=ON -DNPKIT_FLAGS="${npkit_options}" -DCMAKE_INSTALL_PREFIX=$ROCM_PATH -DROCM_PATH=$ROCM_PATH -DONLY_FUNCS="$ONLY_FUNCS" $enable_ninja ../../.
 else
-    CXX=$ROCM_BIN_PATH/hipcc $cmake_executable $cmake_common_options -DBUILD_TESTS=OFF -DNPKIT_FLAGS="${npkit_options}" -DCMAKE_INSTALL_PREFIX=$ROCM_PATH -DROCM_PATH=$ROCM_PATH $enable_ninja ../../.
+    CXX=$ROCM_BIN_PATH/hipcc $cmake_executable $cmake_common_options -DBUILD_TESTS=OFF -DNPKIT_FLAGS="${npkit_options}" -DCMAKE_INSTALL_PREFIX=$ROCM_PATH -DROCM_PATH=$ROCM_PATH -DONLY_FUNCS="$ONLY_FUNCS" $enable_ninja ../../.
 fi
 check_exit_code "$?"
 
