@@ -11,6 +11,7 @@
 #include "nccl.h"
 #include "rccl_bfloat16.h"
 #include "align.h"
+#include "collectives.h"
 #if defined(ENABLE_NPKIT)
 #include "npkit/npkit_struct.h"
 #endif
@@ -20,13 +21,17 @@
 #define NCCL_NUM_FUNCTIONS 5 // SendRecv and AllToAllPivot not included for now
 typedef enum { ncclFuncBroadcast, ncclFuncReduce, ncclFuncAllGather, ncclFuncReduceScatter, ncclFuncAllReduce, ncclFuncSendRecv, ncclFuncSend, ncclFuncRecv, ncclFuncAllToAllPivot, ncclNumFuncs} ncclFunc_t;
 extern const char* ncclFuncStr[NCCL_NUM_FUNCTIONS+2];
+#ifdef ENABLE_COLLTRACE
+extern const char* funcNames[FUNC_INDEX_TOTAL];
+#endif
 
-#define NCCL_NUM_ALGORITHMS 5 // Tree/Ring/CollNet*
+#define NCCL_NUM_ALGORITHMS 6 // Tree/Ring/CollNet*
 #define NCCL_ALGO_TREE 0
 #define NCCL_ALGO_RING 1
 #define NCCL_ALGO_COLLNET_DIRECT 2
 #define NCCL_ALGO_COLLNET_CHAIN 3
 #define NCCL_ALGO_NVLS 4
+#define NCCL_ALGO_NVLS_TREE 5
 extern const char* ncclAlgoStr[NCCL_NUM_ALGORITHMS];
 
 #define NCCL_NUM_PROTOCOLS 3 // Simple/LL/LL128
@@ -54,7 +59,7 @@ union ncclLLFifoLine {
 };
 
 #define WARP_SIZE warpSize
-#define MAXCHANNELS 32
+#define MAXCHANNELS 64
 #define NCCL_MAX_NTHREADS 256
 #define NCCL_SIMPLE_MAX_NTHREADS NCCL_MAX_NTHREADS
 #define NCCL_LL_MAX_NTHREADS NCCL_MAX_NTHREADS
@@ -112,10 +117,10 @@ struct ncclConnInfo {
 };
 
 struct ncclProxyConnector {
-  int rank;
-  int localRank;
+  int tpRank;
+  int tpLocalRank;
+  int sameProcess;
   struct ncclProxyConnection* connection;
-  struct ncclComm* comm;
 };
 
 struct ncclConnector {
@@ -124,7 +129,6 @@ struct ncclConnector {
   struct ncclTransportComm* transportComm;
   void* transportResources;
   struct ncclConnInfo conn;
-  struct ncclComm *comm;
 };
 
 struct ncclRing {
@@ -141,6 +145,9 @@ struct ncclRing {
 };
 
 
+// The root of each tree only has one node down (+1 intra-node).
+#define NCCL_MAX_TREE_ARITY_TOP 2
+// Nodes inside the binary tree can have to two nodes down (+1 intra-node).
 #define NCCL_MAX_TREE_ARITY 3
 struct ncclTree {
   int depth;
@@ -161,18 +168,24 @@ struct ncclDirect {
 
 #define NCCL_CONN_IDX_P2P_NET 2
 #define NCCL_MAX_NVLS_ARITY 8
+#define NCCL_MAX_NVLS_TREE_ARITY 3
 struct ncclNvls {
   int out;
   int nHeads;   // Number of parallel N<->1<->net operations we'll do in parallel; size of up/down
   int headRank; // Index in 0..nHeads-1 I am the head rank of. -1 if I'm not a head rank (no local NIC)
   int up[NCCL_MAX_NVLS_ARITY];
   int down;
+  int treeUp;
+  int treeDown[NCCL_MAX_NVLS_TREE_ARITY];
+  int node;
+  int nNodes;
 };
 
 #define NCCL_MAX_CONNS 3
 struct ncclChannelPeer {
   struct ncclConnector send[NCCL_MAX_CONNS];
   struct ncclConnector recv[NCCL_MAX_CONNS];
+  int refCount;
 };
 
 struct ncclDevComm;
@@ -358,11 +371,16 @@ struct ncclCollTrace {
 };
 static_assert(sizeof(struct ncclCollTrace) == 8*sizeof(int), "ncclCollTrace must have a pow2 size");
 
+union ncclCollTraceTail{
+  uint32_t tail;
+  char padding[4096];
+};
+
 #define COLLTRACE_NUM_ITEMS 8192
 #endif
 
 struct alignas(16) ncclDevChannel {
-  struct ncclDevChannelPeer *peers;
+  struct ncclDevChannelPeer** peers;
   struct ncclRing ring;
   struct ncclTree tree;
   struct ncclTree collnetChain;
@@ -394,7 +412,7 @@ struct ncclDevComm {
 
 #ifdef ENABLE_COLLTRACE
   struct ncclCollTrace* collTrace;
-  volatile uint32_t *collTraceTail;
+  union ncclCollTraceTail *collTraceTail;
   pthread_t collTraceThread;
 #endif
 
@@ -469,4 +487,62 @@ __host__ __device__ constexpr int ncclShmemDynamicSize(int cudaArch = NCCL_CUDA_
   return cudaArch < 700 ? 0 : ncclShmemScratchWarpSize(cudaArch)*(NCCL_MAX_NTHREADS/WARP_SIZE);
 }
 
+// Map the rowIdx to funcIdx
+extern int const ncclFuncRowToId[];
+
+// `ncclFuncIndex()` needs to be in sync with 'ALL_COLLS' in Generate.cmake
+inline int ncclFuncId(int coll, int devRedOp, int type, int algo, int proto) {
+  int row = 0;
+
+  // RING / <all_protos> / Sum / int8_t
+  if (coll == ncclFuncAllGather) {
+    row += proto;
+    goto have_row;
+  }
+  row += NCCL_NUM_PROTOCOLS;
+
+  // <all_algos> / <all_protos> / <all_redops> / <all_types>
+  if (coll == ncclFuncAllReduce) {
+    row += (((algo * NCCL_NUM_PROTOCOLS + proto) * ncclNumDevRedOps + devRedOp) * ncclNumTypes + type) - /*floats for each SumPostDiv*/ 4 * (algo * NCCL_NUM_PROTOCOLS + proto);
+    goto have_row;
+  }
+  row += (NCCL_NUM_ALGORITHMS - 2) * NCCL_NUM_PROTOCOLS * (ncclNumDevRedOps * ncclNumTypes - /*floats for each SumPostDiv*/ 4);
+
+  // RING / SIMPLE / Sum / int8_t
+  if (coll == ncclFuncAllToAllPivot) goto have_row;
+  row += 1;
+
+  // RING / <all_protos> / Sum / int8_t
+  if (coll == ncclFuncBroadcast) {
+    row += proto;
+    goto have_row;
+  }
+  row += NCCL_NUM_PROTOCOLS;
+
+  // RING / <all_protos> / <all_redops> / <all_types>
+  if (coll == ncclFuncReduce) {
+    row += ((proto * ncclNumDevRedOps + devRedOp) * ncclNumTypes + type) - /*floats for each SumPostDiv*/ 4 * proto; 
+    goto have_row;
+  }
+  row += NCCL_NUM_PROTOCOLS * (ncclNumDevRedOps * ncclNumTypes - /*floats for each SumPostDiv*/ 4);
+
+  // RING / <all_protos> / <all_redops> / <all_types>
+  if (coll == ncclFuncReduceScatter) {
+    row += ((proto * ncclNumDevRedOps + devRedOp) * ncclNumTypes + type) - /*floats for each SumPostDiv*/ 4 * proto;
+    goto have_row;
+  }
+  row += NCCL_NUM_PROTOCOLS * (ncclNumDevRedOps * ncclNumTypes - /*floats for each SumPostDiv*/ 4);
+
+  // RING / SIMPLE / Sum / int8_t
+  if (coll == ncclFuncSendRecv) goto have_row;
+  row += 1;
+
+have_row:
+  return ncclFuncRowToId[row];
+}
+
+inline int ncclFuncId_P2p() { return ncclFuncRowToId[FUNC_INDEX_P2P]; }
+
+inline int ncclFuncId_AllToAllPivot() { return ncclFuncRowToId[FUNC_INDEX_ALLTOALL_PIVOT]; }
+
 #endif
