@@ -11,6 +11,7 @@
 #include "msccl/msccl_struct.h"
 
 bool mscclEnabled();
+bool mscclForceEnabled();
 
 void mscclSetIsCallerFlag();
 void mscclClearIsCallerFlag();
@@ -18,6 +19,8 @@ bool mscclIsCaller();
 
 bool mscclAvailable();
 
+ncclResult_t mscclSchedulerInit(ncclComm_t comm, int* numChannelsRequired);
+
 ncclResult_t mscclInit(ncclComm_t comm);
 
 ncclResult_t mscclGroupStart();
