@@ -26,11 +26,22 @@ RCCL_PARAM(MscclEnabled, "MSCCL_ENABLE", 1);
 RCCL_PARAM(MscclForceEnabled, "MSCCL_FORCE_ENABLE", 0);
 static const char* mscclAlgoFilePathEnv = "MSCCL_ALGO_FILE_PATH";
 static std::atomic<bool> mscclInitialized;
-static bool mscclSchedulerTriedLoadAlgo = false;
 static std::mutex mscclLifecycleMutex;
 
 bool mscclEnabled() {
+#ifdef COMPILE_MSCCL_KERNEL
   return rcclParamMscclEnabled();
+#else
+  return false;
+#endif
+}
+
+bool mscclForceEnabled() {
+#ifdef COMPILE_MSCCL_KERNEL
+  return rcclParamMscclForceEnabled();
+#else
+  return false;
+#endif
 }
 
 void mscclSetIsCallerFlag() {
@@ -74,8 +85,21 @@ static const char* mscclUnitTestAlgoDefaultDir = "msccl-unit-test-algorithms";
 static const char* mscclAlgoShareDirPath = "../share/rccl/msccl-algorithms";
 static const char* mscclUnitTestAlgoShareDirPath = "../share/rccl/msccl-unit-test-algorithms";
 
-static ncclResult_t mscclInternalSchedulerInit() {
+static ncclResult_t mscclInternalSchedulerInit(ncclComm_t comm, int* numChannelsRequired) {
+  static bool mscclAlgoMetaLoaded = false;
   mscclStatus& status = mscclGetStatus();
+
+  *numChannelsRequired = 0;
+  // Query numChannelsRequired from loaded algorithm metas
+  if (mscclAlgoMetaLoaded) {
+    for (auto& m : status.algoMetas) {
+      if (comm->nRanks == m.nRanks) {
+        *numChannelsRequired = std::max(*numChannelsRequired, m.nChannels);
+      }
+    }
+    return ncclSuccess;
+  }
+
   const char* mscclAlgoDir = getenv(mscclAlgoDirEnv);
   const char* mscclAlgoShareDir = nullptr;
   std::string mscclAlgoDirStr;
@@ -122,16 +146,28 @@ static ncclResult_t mscclInternalSchedulerInit() {
     fullPath += "/";
     fullPath += entry->d_name;
     NCCLCHECK(mscclGetAlgoMetaFromXmlFile(fullPath.c_str(), &(status.algoMetas.back())));
+    if (status.algoMetas.back().nRanks == comm->nRanks) {
+      *numChannelsRequired = std::max(*numChannelsRequired, status.algoMetas.back().nChannels);
+    }
   }
   if (closedir(dp)) {
     WARN("MSCCL Internal Scheduler: closedir failed, error %d", errno);
     return ncclInvalidUsage;
   }
   status.rankToAlgoHandles.resize(status.algoMetas.size());
+  mscclAlgoMetaLoaded = true;
   return ncclSuccess;
 }
 
-static ncclResult_t mscclSchedulerInit() {
+ncclResult_t mscclSchedulerInit(ncclComm_t comm, int* numChannelsRequired) {
+  *numChannelsRequired = 0;
+  comm->mscclCompatible = mscclCommCompatible(comm);
+  if (!comm->mscclCompatible) {
+    return ncclSuccess;
+  }
+
+  std::lock_guard<std::mutex> lock(mscclLifecycleMutex);
+
   mscclStatus& status = mscclGetStatus();
   bool useInternalScheduler = false;
 
@@ -151,11 +187,14 @@ static ncclResult_t mscclSchedulerInit() {
       useInternalScheduler = true;
     }
   }
+
   if (useInternalScheduler) {
-    NCCLCHECK(mscclInternalSchedulerInit());
+    NCCLCHECK(mscclInternalSchedulerInit(comm, numChannelsRequired));
   } else {
     NCCLCHECK(status.mscclSchedulerPtr->init());
+    *numChannelsRequired = MAXCHANNELS;
   }
+
   return ncclSuccess;
 }
 
@@ -164,28 +203,52 @@ ncclResult_t mscclInit(ncclComm_t comm) {
   mscclThreadLocalStatus threadLocalStatus = mscclGetThreadLocalStatus();
   threadLocalStatus.groupStatus = mscclNoGroup;
   threadLocalStatus.groupDepth = 0;
-  comm->mscclCompatible = mscclCommCompatible(comm);
+  threadLocalStatus.captureId = ULLONG_MAX;
+  threadLocalStatus.captureStatus = mscclNoCapture;
 
   {
     std::lock_guard<std::mutex> lock(mscclLifecycleMutex);
 
+    mscclStatus& status = mscclGetStatus();
+
+    // freeAlgoHandles and needsProxy are initialized globally once and before algorithm pre-processing and connection
+    if (!mscclInitialized.load(std::memory_order_acquire)) {
+      status.freeAlgoHandles.resize(MSCCL_MAX_NUM_ALGOS);
+      for (int i = 0; i < MSCCL_MAX_NUM_ALGOS; i++) {
+        status.freeAlgoHandles[i] = MSCCL_MAX_NUM_ALGOS - i - 1;
+      }
+      status.needsProxy = false;
+    }
+
+    // Pre-process all algorithms for internal scheduler and for different comms.
+    // This is a temp fix to bypass the issue that stream cannot be synchronized during HIP graph capturing,
+    // should use dynamic loading approach after the issue is fixed.
+    if (comm->mscclCompatible && !status.mscclSchedulerPtr) {
+      for (size_t i = 0; i < status.algoMetas.size(); i++) {
+        auto &m = status.algoMetas[i];
+        if (m.nRanks == comm->nRanks) {
+          // Load algorithms
+          if (status.rankToAlgoHandles[i].find(comm->rank) == status.rankToAlgoHandles[i].end()) {
+            NCCLCHECK(mscclLoadAlgo(m.filePath.c_str(), &(status.rankToAlgoHandles[i][comm->rank]), comm->rank));
+          }
+          // Connect algorithms
+          mscclAlgoHandle_t mscclAlgoHandle = status.rankToAlgoHandles[i][comm->rank];
+          if (status.connectedAlgos[comm].find(mscclAlgoHandle) == status.connectedAlgos[comm].end()) {
+            NCCLCHECK(mscclSetupConnections(status.hostAlgos[mscclAlgoHandle], comm));
+            status.connectedAlgos[comm].insert(mscclAlgoHandle);
+          }
+        }
+      }
+    }
+
     if (mscclInitialized.load(std::memory_order_acquire)) {
       return ncclSuccess;
     }
 
-    mscclStatus& status = mscclGetStatus();
-    status.scratchBuffer = nullptr;
-    status.scratchBufferSize = 0;
     status.workIndex = 1;
-    status.freeAlgoHandles.resize(MSCCL_MAX_NUM_ALGOS);
-    for (int i = 0; i < MSCCL_MAX_NUM_ALGOS; i++) {
-      status.freeAlgoHandles[i] = MSCCL_MAX_NUM_ALGOS - i - 1;
-    }
     NCCLCHECK(ncclCudaCalloc(&status.syncFlags, MSCCL_MAX_NUM_THREAD_BLOCKS));
     status.lastStream = nullptr;
-    mscclSchedulerTriedLoadAlgo = false;
-
-    NCCLCHECK(mscclSchedulerInit());
+    NCCLCHECK(mscclInitWorkFifoStatus(&(status.defaultWorkFifoStatus)));
 
     mscclInitialized.store(true, std::memory_order_release);
   }
@@ -207,6 +270,11 @@ static ncclResult_t mscclInternalSchedulerSelectAlgo(struct mscclSchedulerParam*
   mscclStatus& status = mscclGetStatus();
   param->scheduled = false;
 
+  // Current MSCCL doesn't support pre/post op
+  if (param->op >= ncclAvg) {
+    return ncclSuccess;
+  }
+
   // Whether the algorithm is in-place
   bool isInPlace = false;
   if (param->func == mscclFuncReduce ||
@@ -234,12 +302,6 @@ static ncclResult_t mscclInternalSchedulerSelectAlgo(struct mscclSchedulerParam*
         m.nRanks == param->nRanks &&
         m.func == param->func &&
         (isInPlace ? m.inPlace : m.outOfPlace)) {
-      // If not loaded for current rank, load it
-      if (status.rankToAlgoHandles[i].find(param->rank) == status.rankToAlgoHandles[i].end()) {
-        mscclAlgoHandle_t algoHandle;
-        NCCLCHECK(mscclLoadAlgo(m.filePath.c_str(), &algoHandle, param->rank));
-        status.rankToAlgoHandles[i][param->rank] = algoHandle;
-      }
       param->handle = status.rankToAlgoHandles[i][param->rank];
       param->scheduled = true;
       return ncclSuccess;
@@ -254,7 +316,8 @@ static ncclResult_t mscclSchedulerSelectAlgo(struct mscclSavedSchedulerParam* pa
   if (status.mscclSchedulerPtr) {
     NCCLCHECK(status.mscclSchedulerPtr->selectAlgo(&(param->p)));
   } else {
-    if (param->comm->topo->mscclEnabled || rcclParamMscclForceEnabled()) {
+    // Disable MSCCL algorithms if machine type is not matching
+    if (param->comm->topo->mscclEnabled || mscclForceEnabled()) {
       NCCLCHECK(mscclInternalSchedulerSelectAlgo(&(param->p)));
     } else {
       param->p.scheduled = false;
@@ -285,6 +348,7 @@ static ncclResult_t mscclSetSavedSchedulerParam(
   param->p.nRanks = comm->nRanks;
   param->comm = comm;
   param->stream = stream;
+  param->p.opCount = comm->opCount;
   return ncclSuccess;
 }
 
@@ -302,9 +366,27 @@ static ncclResult_t mscclSaveCountsAndDispls(struct mscclSavedSchedulerParam* pa
   return ncclSuccess;
 }
 
+const char *mscclFuncNames[] = {
+            "mscclFuncReduce",
+            "mscclFuncBroadcast",
+            "mscclFuncAllReduce",
+            "mscclFuncReduceScatter",
+            "mscclFuncAllGather",
+            "mscclFuncSend",
+            "mscclFuncRecv",
+            "mscclFuncGather",
+            "mscclFuncScatter",
+            "mscclFuncAllToAll",
+            "mscclFuncAllToAllv",
+          };
+
 static ncclResult_t mscclRunSavedParams() {
   mscclThreadLocalStatus& threadLocalStatus = mscclGetThreadLocalStatus();
   for (auto& param : threadLocalStatus.savedSchedulerParams) {
+    INFO(NCCL_COLL,"%s: opCount %lx sendbuff %p recvbuff %p count %zi datatype %d op %d root %d comm %p [nranks=%d] stream %p",
+    mscclFuncNames[param.p.func], param.p.opCount, param.p.sendBuff, param.p.recvBuff, param.p.count,
+    param.p.dataType, param.p.op, param.p.root, param.comm, param.p.nRanks, param.stream);
+
     NCCLCHECK(mscclRunAlgo(
       param.p.sendBuff, param.p.sendCounts, param.p.sDisPls,
       param.p.recvBuff, param.p.recvCounts, param.p.rDisPls,
@@ -381,8 +463,6 @@ ncclResult_t mscclEnqueueCheck(
     size_t count, ncclDataType_t dataType, int root, int peer, ncclRedOp_t op,
     mscclFunc_t func, ncclComm_t comm, hipStream_t stream) {
   mscclThreadLocalStatus& threadLocalStatus = mscclGetThreadLocalStatus();
-  hipStreamCaptureStatus captureStatus;
-  unsigned long long pid;
 
   threadLocalStatus.savedSchedulerParams.push_back({});
   NCCLCHECK(mscclSetSavedSchedulerParam(
@@ -393,29 +473,16 @@ ncclResult_t mscclEnqueueCheck(
   switch (threadLocalStatus.groupStatus) {
     case mscclNoGroup:
       if (comm->mscclCompatible) {
-        if (stream == (hipStream_t)0) {
-          captureStatus = hipStreamCaptureStatusNone;
-        } else {
-          CUDACHECK(hipStreamGetCaptureInfo(stream, &captureStatus, &pid));
-        }
-        if (captureStatus == hipStreamCaptureStatusNone) {
           NCCLCHECK(mscclSchedulerSelectAlgo(&threadLocalStatus.savedSchedulerParams.back()));
           if (threadLocalStatus.savedSchedulerParams.back().p.scheduled) {
             NCCLCHECK(mscclRunSavedParams());
             break;
           }
         }
-      }
       NCCLCHECK(mscclFallBackSavedParams());
       break;
     case mscclGroupSupportedOp:
       if (comm->mscclCompatible) {
-        if (stream == (hipStream_t)0) {
-          captureStatus = hipStreamCaptureStatusNone;
-        } else {
-          CUDACHECK(hipStreamGetCaptureInfo(stream, &captureStatus, &pid));
-        }
-        if (captureStatus == hipStreamCaptureStatusNone) {
           NCCLCHECK(mscclSchedulerSelectAlgo(&threadLocalStatus.savedSchedulerParams.back()));
           if (threadLocalStatus.savedSchedulerParams.back().p.scheduled) {
             // Only save counts and displs when there is suitable MSCCL algorithm for this
@@ -423,8 +490,8 @@ ncclResult_t mscclEnqueueCheck(
             break;
           }
         }
-      }
       threadLocalStatus.groupStatus = mscclGroupUnsupportedOp;
+      NCCLCHECK(mscclFallBackSavedParams());
     case mscclGroupUnsupportedOp:
       NCCLCHECK(mscclFallBackSavedParams());
       break;
@@ -481,13 +548,14 @@ ncclResult_t mscclTeardown() {
     for (auto &p : status.devAlgos) {
       CUDACHECK(hipFree(p.second));
     }
-    CUDACHECK(hipFree(status.scratchBuffer));
     CUDACHECK(hipFree(status.syncFlags));
     status.hostAlgos.clear();
     status.devAlgos.clear();
     status.freeAlgoHandles.clear();
-    status.scratchBuffer = nullptr;
-    status.scratchBufferSize = 0;
+    for (auto &p : status.scratchBuffers) {
+      CUDACHECK(hipFree(p.second));
+    }
+    status.scratchBuffers.clear();
     status.connectedAlgos.clear();
     if (status.mscclSchedulerPtr) {
       NCCLCHECK(status.mscclSchedulerPtr->teardown());
@@ -497,6 +565,10 @@ ncclResult_t mscclTeardown() {
     } else {
       NCCLCHECK(mscclInternalSchedulerTeardown());
     }
+    NCCLCHECK(mscclDestroyWorkFifoStatus(&(status.defaultWorkFifoStatus)));
+    for (auto &p : status.graphWorkFifoStatus) {
+      NCCLCHECK(mscclDestroyWorkFifoStatus(&(p.second)));
+    }
     mscclInitialized.store(false, std::memory_order_release);
   }
 
