@@ -4,6 +4,7 @@
  ************************************************************************/
 
 #include "msccl/msccl_status.h"
+#include "msccl/msccl_struct.h"
 
 mscclStatus& mscclGetStatus() {
   static mscclStatus status;
@@ -14,3 +15,8 @@ mscclThreadLocalStatus& mscclGetThreadLocalStatus() {
   static thread_local mscclThreadLocalStatus threadLocalStatus;
   return threadLocalStatus;
 }
+
+mscclSavedProxyArgs& mscclGetSavedProxyArgs() {
+  static mscclSavedProxyArgs savedProxyArgs;
+  return savedProxyArgs;
+}
